#!/usr/bin/python
# -*- coding: utf-8 -*-

DEFAULT_WAIT = 1 # tempo em segundos de espera entre partidas

import os, time

import pygame
from pygame.locals import *
from pygame.sprite import RenderUpdates
from pygame.time   import Clock
from pygame.event  import Event
from pygame.color  import Color
from pygame.font   import Font


from paddle       import Paddle
from ball         import Ball
from scoredisplay import ScoreDisplay

from SimpleXMLRPCServer import SimpleXMLRPCServer
from xmlrpclib          import ServerProxy

class Game:
    screen      = None
    run         = True
    img_p1      = None
    img_p2      = None
    img_ball    = None
    img_table   = None
    wait        = DEFAULT_WAIT
    dt          = 16
    r_event     = None
    catch_from  = 0

    
    def __init__( self, size, fullscreen, server, client ):
        """
        Esta é a função que inicializa o pygame, define a resolução da tela,
        caption, e disabilitamos o mouse dentro desta.
        """

        if not ( server or client ):
            raise Exception( "Precisa ser cliente ou servidor!!!" )
        
        pygame.init()
        self.flags = DOUBLEBUF
        if fullscreen:
            self.flags |= FULLSCREEN
        self.screen      = pygame.display.set_mode( size, self.flags )
        self.screen_size = self.screen.get_size()
        self.client      = client
        self.server      = server
        self.r_event     = []
        
        pygame.mouse.set_visible( 0 )
        pygame.display.set_caption( 'Ping Pong -- NET' )        

        self.load_images()
    # init()



    def load_images( self ):
        """
        Lê as imagens necessarias pelo jogo.
        """
        p = "imagens" + os.path.sep
        self.img_p1    = pygame.image.load( p + "p1.png" )
        self.img_p2    = pygame.image.load( p + "p2.png" )
        self.img_ball  = pygame.image.load( p + "ball.png" )
        self.img_table = pygame.image.load( p + "table.png" )
    # load_images()



    def setup( self ):
        """
        Faz a configuração inicial, imprime o fundo e inicia os jogadores.
        """
        # Imprimimos o fundo. Ele não precisa ser um personagem pois ele
        # não atua (muda, move, ...)
        self.screen.blit( self.img_table, ( 0, 0 ) )
        pygame.display.flip()
        
        # Area de movimentacao dentro da mesa (x, y, w, h)
        tma = Rect( 10, 30, 780, 570 )

        # Area de movimentacao de cada jogador:
        p1_ma   = Rect( 0, tma.top, tma.left, tma.height )
        p2_ma   = Rect( tma.left + tma.width, tma.top, tma.left, tma.height )
        ball_ma = Rect( 0, tma.top, tma.width + 2 * tma.left, tma.height  )
        

        
        # Inicia Jogadores: imagem, posicao inicial, area de movimentacao
        s = os.path.join( "sounds", "paddle.wav" )
        self.p1   = Paddle( self.img_p1, p1_ma.topleft, p1_ma, s )
        self.p2   = Paddle( self.img_p2, p2_ma.topleft, p2_ma, s )
        self.ball = Ball( self.img_ball, ball_ma, tma )

        # Centraliza Jogadores:
        self.p1.rect.center = p1_ma.center
        self.p2.rect.center = p2_ma.center        

        # Grupo de elementos atuantes:
        self.group = RenderUpdates( [ self.ball,
                                      self.p1,
                                      self.p2 ] )

        
        # Placar:
        r = Rect( 0, 0, tma.width + 2 * tma.left, tma.top )
        self.score = ScoreDisplay( r )
    # setup()

    
    def handle_events( self ):
        """
        Trata o evento e toma a ação necessária.
        """
        if self.server is not None:
            remote = self.r_event
            local  = []
            for event in pygame.event.get():
                local += [ event ]
            
            self.r_event = []
            for p, events in ( ( self.p1, local ), ( self.p2, remote ) ):
                for event in events:
                    t = event.type
                    if t in ( KEYDOWN, KEYUP ):
                        k = event.key

                    if t == QUIT:
                        self.run = False

                    elif t == KEYDOWN:
                        if   k == K_ESCAPE:
                            self.run = False
                        elif k == K_UP:                            
                            p.set_move_up()
                        elif k == K_DOWN:
                            p.set_move_down()

                    elif t == KEYUP:
                        if k == K_UP:
                            p.unset_move_up()
                        elif k == K_DOWN:
                            p.unset_move_down()
                
            self.r_event = []
            
        elif self.client is not None:
            events = []
            for e in pygame.event.get():
                # Não é possivel transferir um Event()
                # então transmita só o interessante:
                k = 0
                try:
                    k = e.key
                except:
                    k = 0
                events += [ ( e.type, { "key": k }) ]
                
            self.client.input_event( events )
    # handle_events()



    def manage( self ):
        """
        Verifica se a bola saiu e algum jogador pegou e faça o necessário.

        Aqui é que podemos adicionar mais lógica no jogo.
        """
        self.catch_from = 0
        out = self.ball.check_out()
        if out:
            if out == -1:
                if self.p1.catch( self.ball ):
                    # P1 pegou, retorne a bola
                    self.catch_from = -1
                    # inverte trajetória da bola
                    self.ball.invert_speed_x()
                    # Transfere movimento vertical para bola ( 50 % )
                    self.ball.inc_speed_y( self.p1.get_speed()[ 1 ] * 0.5 )
                    return
                else:
                    # Ponto para P2
                    self.score.p2 += 1
                    self.restart()
            
            if out == 1:
                if self.p2.catch( self.ball ):
                    # P2 pegou, retorne a bola
                    self.catch_from = 1
                    # inverte trajetória da bola
                    self.ball.invert_speed_x()
                    # Transfere movimento vertical para bola ( 50 % )
                    self.ball.inc_speed_y( self.p2.get_speed()[ 1 ] * 0.5 )
                    return
                else:
                    # Ponto para P1
                    self.score.p1 += 1
                    self.restart()

        if self.server:
            self.server.handle_request() # input_event()
            self.server.handle_request() # get_state()
    # manage()



    def restart( self ):
        """Reinicia a jogada."""
        self.ball.start()
        self.wait = DEFAULT_WAIT
    # restart()



    def do_wait( self ):
        """Verifica se deve esperar e configura as ações."""
        if self.wait > 0:
            # decremente o tempo a esperar
            self.wait -= float( self.dt ) / 1000
            # pause a bola
            self.ball.pause = True

            if self.wait <= 0:
                # Já expirou o tempo, volte ao estado normal
                self.wait = 0
                self.ball.pause = False
    # do_wait()



    def show_msg( self, msg, fgcolor="0xffffff", bgcolor="0x444444",
                  timeout=0.5):
        """
        Mostra mensagem na tela
        """
        f = Font( None, 30 )
        i = f.render( msg, True, Color( fgcolor ), Color( bgcolor ) )
        r = i.get_rect()
        r.center = self.screen.get_rect().center

        bkp = self.screen.convert()
        self.screen.blit( i, r )
        pygame.display.flip()

        if timeout > 0:
            time.sleep( timeout )
            self.screen.blit( bkp, ( 0, 0 ) )
            pygame.display.flip()
    # show_msg()



    def server_init( self ):
        """
        Inicia servidor.
        """
        if isinstance( self.server, ( tuple, list ) ):
            self.server = SimpleXMLRPCServer( tuple( self.server ) )
            self.server.register_function( self.get_state )
            self.server.register_function( self.input_event )

            self.show_msg( u"Esperando cliente..." )
            self.server.handle_request()
    # server_init()



    def client_init( self ):
        """
        Conecta ao servidor.
        """
        if isinstance( self.client, ( str, unicode ) ):
            url = self.client
            self.client = ServerProxy( self.client )
            self.ball.dumb = True
            while True:
                try:
                    self.show_msg( u"Tentando conexão a %s." % url )
                    self.client.get_state()
                    break
                except Exception, e:
                    self.show_msg( u"Erro, tentando novamente." )
    # client_init()



    def client_quit( self ):
        """
        Avisa que cliente saiu.
        """
        self.client.input_event( [ ( QUIT, {} ) ] )
    # client_quit()



    def get_state( self ):
        """
        Retorna o estado dos elementos:
           [ ( posicao_p1, pontos_p1 ),
             ( posicao_p2, pontos_p2 ),
             ( posicao_ball, quem_pegou_a_bola ) ]

        Para uso pelo cliente
        """
        if self.run:
            state = [ ( self.p1.rect.topleft, self.score.p1 ),
                      ( self.p2.rect.topleft, self.score.p2 ),
                      ( self.ball.rect.topleft, self.catch_from ) ]
        else:
            state = 0
        return state
    # get_state()



    def update_state( self ):
        """
        Atualiza estado baseado no seridor.
        """
        if self.client is not None:
            state = self.client.get_state()
            if state == 0:
                self.run = False
            else:
                self.p1.rect.topleft = state[ 0 ][ 0 ]
                self.score.p1        = state[ 0 ][ 1 ]

                self.p2.rect.topleft = state[ 1 ][ 0 ]
                self.score.p2        = state[ 1 ][ 1 ]

                self.ball.rect.topleft = state[ 2 ][ 0 ]
                self.catch_from        = state[ 2 ][ 1 ]
    # update_state()



    def input_event( self, event ):
        """Envia lista de eventos do cliente para servidor."""
        if isinstance( event, ( tuple, list ) ):
            for e in event:
                self.r_event += [ Event( e[ 0 ],  e[ 1 ] ) ]
        return True
    # input_event()
    
            
    
    def loop( self ):
        """
        Laço principal
        """
        # Faça as configurações iniciais.
        self.setup()

        if self.client:
            self.client_init()
            self.server = None
        else:
            self.server_init()
            self.client = None
        
        # Inicializamos o relogio e o dt que vai limitar o valor de
        # frames por segundo do jogo
        clock         = Clock()
        interval      = 0

        # assim iniciamos o loop principal do programa
        while self.run:
            clock.tick( 1000 / self.dt )
            interval += 1
            if interval > 1000 and self.dt > 1:
                self.dt -= 1
                interval = 0

            # Trabalha eventos
            self.handle_events()

            # Verifica se deve esperar
            self.do_wait()

            # Atualiza estado baseado no servidor
            self.update_state()

            # Atualiza posicoes
            self.group.update( self.dt )

            # Faz manutenção do jogo, verifica se alguém perdeu, ...
            self.manage()

            # Limpa posicao atual de elementos, para nao deixar um rastro
            self.group.clear( self.screen, self.img_table )

            # Desenha elementos, guarda areas atualizadas
            r = self.group.draw( self.screen )

            # Atualiza o display com as areas modificadas
            pygame.display.update( r )

            # Toca som correspondente
            if   self.catch_from == -1:
                self.p1.soundeffect()
            elif self.catch_from ==  1:
                self.p2.soundeffect()
                

            # Mostra Placar e atualiza:
            self.score.draw( self.screen )
            pygame.display.update( self.score )

            # ao fim do desenho temos que trocar o front buffer e o back buffer
            # Somente use se usar double buffer!!!
            if self.flags & DOUBLEBUF:
                pygame.display.flip()

            #print "FPS: %0.2f" % clock.get_fps()
        # while self.run

        self.show_msg( "Jogo terminado. Avisando o outro jogador." )
        
        if self.client is not None:
            # Avisa que cliente saiu
            self.client_quit()
            
        elif self.server is not None:
            # Espera cliente para avisar que o servidor saiu.
            self.server.handle_request() 

        self.show_msg( "FIM. Pressione qualquer tecla.", timeout=-1 )
        run = True
        while run:
            for e in pygame.event.get():
                if e.type in ( QUIT, KEYDOWN ):
                    run = False
    # loop()
# Game
